/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2015-2020 Amazon.com, Inc. or its affiliates.
 * All rights reserved.
 */
#define	ENA_GEN_DATE	"Fri Sep 18 17:09:00 IDT 2020"
#define	ENA_GEN_COMMIT	"0f80d82"
