/*
 * https://www.boost.org/doc/libs/1_82_0/libs/url/doc/html/url/overview.html#url.overview.requirements
 *
 * To use the library as header-only; that is, to eliminate the requirement 
 * to link a program to a static or dynamic Boost.URL library, 
 * simply place the following line in exactly one source file in your project.
 */
#include <boost/url/src.hpp>
